/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.utils;

import static org.eclipse.emf.ecore.util.EcoreUtil.delete;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.copy;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.fortiss.tooling.base.model.base.ConnectorBase;
import org.fortiss.tooling.base.model.base.HierarchicElementBase;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.element.IModelElementReference;
import org.fortiss.tooling.base.model.element.IModelElementSpecification;
import org.fortiss.tooling.base.model.layout.ILayoutData;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;
import org.fortiss.tooling.kernel.model.INamedElement;

/**
 * Utility methods for implementation of drop-in replacement migrations of (sub-types of) types
 * defined in {@code org.fortiss.tooling.base}.
 *
 * @author barner
 */
public class MigrationUtils { // NO_UCD
	// Migrators are retired after a release, but these utilities should be available when needed.
	// See https://af3-developer.fortiss.org/projects/autofocus3/wiki/Model_Migration

	/**
	 * Migrates an {@code oldConnectorBase} (of an (old) concrete sub-type of {@link ConnectorBase})
	 * to a {@code newConnectorBase} (of another (new) concrete sub-type of {@link ConnectorBase} ),
	 * including all relevant attributes and children. Side effect: This method removes the
	 * {@code oldConnectorBase} from the model and deletes it.
	 * 
	 * @param oldConnectorBase
	 *            Old {@link ConnectorBase} to be migrated. Side effect: Removed from the model and
	 *            deleted.
	 * @param newConnectorBase
	 *            New {@link ConnectorBase} which should replace {@code oldConnectorBase}. It must
	 *            be created by the caller of this method.
	 */
	public static void migrateConnectorBase(ConnectorBase oldConnectorBase,
			ConnectorBase newConnectorBase) {

		migrateINamedCommentedElement(oldConnectorBase, newConnectorBase);
		migrateILayoutedModelElement(oldConnectorBase, newConnectorBase);
		migrateIModelElement(oldConnectorBase, newConnectorBase);
		migrateIConnector(oldConnectorBase, newConnectorBase);

		delete(oldConnectorBase);
	}

	/**
	 * Migrates an {@code oldConnectorBase} (of an (old) concrete sub-type of {@link ConnectorBase})
	 * to a {@code newConnectorBase} (of another (new) concrete sub-type of {@link ConnectorBase} )
	 * that is owned by the given new {@link IHierarchicElement},
	 * including all relevant attributes and children.
	 * 
	 * @param oldConnectorBase
	 *            Old {@link ConnectorBase} to be migrated. Side effect: Removed from the model and
	 *            deleted.
	 * @param newConnectorBase
	 *            New {@link ConnectorBase} which should replace {@code oldConnectorBase}. It must
	 *            be created by the caller of this method.
	 * @param newElement
	 *            Element that will own the {@code newConnectorBase}.
	 */
	public static void migrateConnectorBase(ConnectorBase oldConnectorBase,
			ConnectorBase newConnectorBase, IHierarchicElement newElement) {

		migrateINamedCommentedElement(oldConnectorBase, newConnectorBase);
		migrateILayoutedModelElement(oldConnectorBase, newConnectorBase);
		migrateIModelElement(oldConnectorBase, newConnectorBase);
		newElement.getConnectors().add(newConnectorBase);
	}

	/**
	 * Migrates all properties defined in the {@link IConnector} interface from
	 * {@code oldIConnector} to {@code newIConnector}.
	 * 
	 * @param oldIConnector
	 *            Old {@link IConnector} to be migrated. Side effect: {@code target} and
	 *            {@code source} are reset, and the {@code oldIConnector} is removed from its owner.
	 * @param newIConnector
	 *            Replacement {@link IConnector}. Side effect: {@code target} and {@code source} are
	 *            set, and it is replaces {@code oldIConnector} in the connector list of the
	 *            corresponding owner model element.
	 */
	public static void migrateIConnector(IConnector oldIConnector, IConnector newIConnector) {
		EList<IConnection> connections = new BasicEList<IConnection>(oldIConnector.getIncoming());
		for(IConnection connection : connections) {
			connection.setTarget(newIConnector);
		}

		connections = new BasicEList<IConnection>(oldIConnector.getOutgoing());
		for(IConnection connection : connections) {
			connection.setSource(newIConnector);
		}

		// Insert at the same position in the owner's connector list
		EList<IConnector> connectors = oldIConnector.getOwner().getConnectors();
		int i = connectors.indexOf(oldIConnector);
		connectors.remove(i);
		connectors.add(i, newIConnector);
	}

	/**
	 * Migrates all properties defined in the {@link IModelElement} interface from
	 * {@code oldIModelElement} to {@code newIModelElement}.
	 * 
	 * @param oldIModelElement
	 *            Old {@link IModelElement} to be migrated.
	 * @param newIModelElement
	 *            Replacement {@link IModelElement}. Copies of {@link IModelElementSpecification}s
	 *            and {@link IModelElementReference}s of {@code oldIModelElement} are
	 *            added.
	 */
	public static void migrateIModelElement(IModelElement oldIModelElement,
			IModelElement newIModelElement) {
		newIModelElement.getSpecifications().clear();
		for(IModelElementSpecification specification : oldIModelElement.getSpecifications()) {
			newIModelElement.getSpecifications().add(copy(specification));
		}

		newIModelElement.getReferencedBy().clear();
		for(IModelElementReference reference : oldIModelElement.getReferencedBy()) {
			newIModelElement.getReferencedBy().add(copy(reference));
		}
	}

	/**
	 * Migrates all properties defined in the {@link ILayoutedModelElement} interface from
	 * {@code oldIConnector} to {@code newIConnector}.
	 * 
	 * @param oldILayoutedModelElement
	 *            Old {@link ILayoutedModelElement} to be migrated.
	 * @param newILayoutedModelElement
	 *            Replacement {@link ILayoutedModelElement}. Copies of {@link ILayoutData} elements
	 *            of {@code oldLayoutedModelElement} are added.
	 */
	public static void migrateILayoutedModelElement(ILayoutedModelElement oldILayoutedModelElement,
			ILayoutedModelElement newILayoutedModelElement) {
		newILayoutedModelElement.getLayoutData().clear();
		for(ILayoutData layoutData : oldILayoutedModelElement.getLayoutData()) {
			newILayoutedModelElement.getLayoutData().add(copy(layoutData));
		}
	}

	/**
	 * Migrates all properties defined in the {@link INamedElement} interface from
	 * {@code oldINamedElement} to {@code newINamedElement}.
	 * 
	 * @param oldINamedElement
	 *            Old {@link INamedElement} to be migrated.
	 * @param newINamedElement
	 *            Replacement {@link INamedElement}.
	 */
	public static void migrateINamedElement(INamedElement oldINamedElement,
			INamedElement newINamedElement) {
		newINamedElement.setName(oldINamedElement.getName());
		newINamedElement.setId(oldINamedElement.getId());
	}

	/**
	 * Migrates all properties defined in the {@link INamedCommentedElement} interface from
	 * {@code oldINamedCommentedElement} to {@code newINamedCommentedElement}.
	 * 
	 * @param oldINamedCommentedElement
	 *            Old {@link INamedCommentedElement} to be migrated.
	 * @param newINamedCommentedElement
	 *            Replacement {@link INamedCommentedElement}.
	 */
	public static void migrateINamedCommentedElement(
			INamedCommentedElement oldINamedCommentedElement,
			INamedCommentedElement newINamedCommentedElement) {
		migrateINamedElement(oldINamedCommentedElement, newINamedCommentedElement);
		newINamedCommentedElement.setComment(oldINamedCommentedElement.getComment());
	}

	/** Migrates an {@link HierarchicElementBase} to an new {@link HierarchicElementBase}. */
	public static void migrateHierarchicElementBase(HierarchicElementBase oldElement,
			HierarchicElementBase newElement) {
		migrateIModelElement(oldElement, newElement);
		migrateILayoutedModelElement(oldElement, newElement);
		migrateINamedCommentedElement(oldElement, newElement);
	}

	/**
	 * Migrates all properties defined in the {@link IAnnotatedSpecification} interface from
	 * {@code oldIAnnotatedSpecification} to {@code newIAnnotatedSpecification}.
	 * <p>
	 * <b>NOTE</b>: After calling this method the annotation is already added to the list of
	 * specifications of the corresponding {@link IModelElement}. More precisely:
	 * {@code element.getSpecifications().add(annotation)} must NOT be called.
	 * 
	 * @param oldIAnnotatedSpecification
	 *            Old {@link INamedCommentedElement} to be migrated.
	 * @param newIAnnotatedSpecification
	 *            Replacement {@link INamedCommentedElement}.
	 */
	public static void migrateIAnnotatedSpecification(
			IAnnotatedSpecification oldIAnnotatedSpecification,
			IAnnotatedSpecification newIAnnotatedSpecification) {
		migrateINamedElement(oldIAnnotatedSpecification, newIAnnotatedSpecification);
		newIAnnotatedSpecification
				.setSpecificationOf(oldIAnnotatedSpecification.getSpecificationOf());
	}
}
