/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.spiderchart.widget;

import static org.fortiss.tooling.spiderchart.util.AxisUtils.getSegmentPoint;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Path;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Display;
import org.fortiss.tooling.base.ui.utils.GCStateManager;
import org.fortiss.tooling.spiderchart.model.AxisBase;
import org.fortiss.tooling.spiderchart.model.DataSeries;
import org.fortiss.tooling.spiderchart.model.DoubleAxis;
import org.fortiss.tooling.spiderchart.model.EnumerationAxis;
import org.fortiss.tooling.spiderchart.model.SpiderChart;
import org.fortiss.tooling.spiderchart.style.AxisStyle;
import org.fortiss.tooling.spiderchart.style.ChartStyle;
import org.fortiss.tooling.spiderchart.style.DataSeriesStyle;
import org.fortiss.tooling.spiderchart.style.FillStyle;
import org.fortiss.tooling.spiderchart.style.FontStyle;

/**
 * Class for drawing the spider chart.
 *
 * @author mondal
 * @author hoelzl
 */
public final class SpiderChartWidget extends SpiderChartWidgetBase {
	/** The center point of the chart. */
	private Point center = new Point(0, 0);
	/** The end points of the axes. */
	private Map<AxisBase, Point> axesEnds = new HashMap<AxisBase, Point>();
	/** The angle in degree between two axes. */
	private double segmentDegree = 0.0;

	/** Constructor. */
	public SpiderChartWidget(SpiderChart chart, ChartStyle style) {
		super(chart, style);
	}

	/** Draws the spider chart. */
	public void draw(GCStateManager gcState) {
		if(!style.isUseIndividualAxisSegments()) {
			drawBackground(gcState);
			drawWebLines(gcState);
		}
		drawAxes(gcState);
		for(DataSeries data : chart.getDataSeries()) {
			drawDataSeries(gcState, data);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void setBounds(int x, int y, int w, int h) {
		super.setBounds(x, y, w, h);
		computeAxisEndPoints();
	}

	/** Computes the end points of the axes. */
	private void computeAxisEndPoints() {
		// compute the axis points
		List<AxisBase> axes = chart.getAxes();
		int size = axes.size();
		segmentDegree = 360.0 / size;
		double currentAngleDegree = style.getStartAngleDegree();

		int halfWidth = getWidth() / 2;
		int halfHeight = getHeight() / 2;
		center.x = getX() + halfWidth;
		center.y = getY() + halfHeight;

		double relativeMargin = 1.0 - style.getRelativeMargin();
		int wExtent = (int)(relativeMargin * halfWidth);
		int hExtent = (int)(relativeMargin * halfHeight);
		int extent = Math.min(wExtent, hExtent);

		for(int i = 0; i < size; i++) {
			AxisBase axis = axes.get(i);
			// compute end point
			double startAngleRad = Math.PI * currentAngleDegree / 180.0;
			int extentX = (int)(Math.cos(startAngleRad) * extent);
			int extentY = (int)(Math.sin(startAngleRad) * extent);
			Point outer = new Point(center.x + extentX, center.y + extentY);
			axesEnds.put(axis, outer);
			currentAngleDegree = (currentAngleDegree + segmentDegree) % 360.0;
		}
	}

	/** Draws the background if style requires all axes to have equal number of segments. */
	private void drawBackground(GCStateManager gcState) {
		Path path = new Path(Display.getCurrent());
		boolean first = true;
		for(AxisBase axis : chart.getAxes()) {
			Point p = axesEnds.get(axis);
			if(first) {
				path.moveTo(p.x, p.y);
				first = false;
			} else {
				path.lineTo(p.x, p.y);
			}
		}
		path.close();

		gcState.backgroundStoreAndCreate(style.getBackgroundFillStyle().getRGBColor());
		gcState.GC().fillPath(path);
		gcState.backgroundRestoreAndDispose();

		path.dispose();
	}

	/** Draws the web lines of the background if all axes have equal number of segments. */
	private void drawWebLines(GCStateManager gcState) {
		gcState.foregroundStoreAndCreate(style.getBackgroundLineStyle().getRGBColor());
		for(int i = 1; i <= style.getAxisSegments(); i++) {
			Path p = computePath(i);
			gcState.GC().drawPath(p);
			p.dispose();
		}
		gcState.foregroundRestoreAndDispose();
	}

	/** Computes the i-th web line path. */
	private Path computePath(int i) {
		double ratio = (double)i / (double)style.getAxisSegments();
		boolean first = true;
		Path p = new Path(Display.getCurrent());
		for(AxisBase axis : chart.getAxes()) {
			Point outer = axesEnds.get(axis);
			int xExtent = outer.x - center.x;
			int yExtent = outer.y - center.y;
			int x = center.x + (int)(ratio * xExtent);
			int y = center.y + (int)(ratio * yExtent);
			if(first) {
				p.moveTo(x, y);
				first = false;
			} else {
				p.lineTo(x, y);
			}
		}
		p.close();
		return p;
	}

	/** Draws the data series. */
	private void drawDataSeries(GCStateManager gcState, DataSeries data) {
		DataSeriesStyle dStyle = style.getDataSeriesStyle(data);
		Path path = new Path(Display.getCurrent());
		boolean first = true;
		for(AxisBase axis : chart.getAxes()) {
			Point p = computeDataPoint(axis, data);
			if(first) {
				path.moveTo(p.x, p.y);
				first = false;
			} else {
				path.lineTo(p.x, p.y);
			}
			if(dStyle.isShowIndicators()) {
				drawIndicator(gcState, dStyle, p);
			}
			if(dStyle.isShowIndicatorLabels()) {
				drawIndicatorLabel(gcState, dStyle, p, axis, data.getAxisValue(axis));
			}
		}
		path.close();

		// draw the filled area with alpha value.
		FillStyle fStyle = dStyle.getFillStyle();
		gcState.alphaStoreAndSet(fStyle.getAlpha());
		gcState.backgroundStoreAndCreate(fStyle.getRGBColor());
		gcState.GC().fillPath(path);
		gcState.backgroundRestoreAndDispose().alphaRestore();

		// draw the outline without alpha value
		gcState.foregroundStoreAndCreate(dStyle.getLineStyle().getRGBColor());
		gcState.GC().drawPath(path);
		gcState.foregroundRestoreAndDispose();

		path.dispose();
	}

	/** Draws the indicator labels next to the chart point. */
	private void drawIndicatorLabel(GCStateManager gcState, DataSeriesStyle dStyle, Point p,
			AxisBase axis, Object value) {
		gcState.fontStoreAndCreate(dStyle.getIndicatorLabelStyle().getFontData());
		gcState.foregroundStoreAndCreate(dStyle.getIndicatorLabelStyle().getRGBColor());

		GC gc = gcState.GC();
		String label = dStyle.getFormattedLabel(axis, value);
		Point textExtent = gc.textExtent(label);
		// data series indicator labels are always to the right of the point
		int x = p.x + 4;
		int y = p.y;
		if(y > center.y) {
			y = y + 4;
		} else if(y < center.y) {
			y = y - 4 - textExtent.y;
		} else {
			y = y - textExtent.y / 2;
		}
		gc.drawString(label, x, y, true);

		gcState.foregroundRestoreAndDispose().fontRestoreAndDispose();
	}

	/** Draws the filled indicator rectangle. */
	private void drawIndicator(GCStateManager gcState, DataSeriesStyle dStyle, Point p) {
		int size = dStyle.getIndicatorSize();
		gcState.backgroundStoreAndCreate(dStyle.getLineStyle().getRGBColor());
		gcState.GC().fillRectangle(p.x - size / 2, p.y - size / 2, size, size);
		gcState.backgroundRestoreAndDispose();
	}

	/** Computes the point on the given axis for the given data. */
	@SuppressWarnings("unchecked")
	private Point computeDataPoint(AxisBase axis, DataSeries data) {
		Double ratio = 0.0;
		if(axis instanceof DoubleAxis) {
			DoubleAxis da = (DoubleAxis)axis;
			ratio = da.getAxisRatio((Double)data.getAxisValue(axis));
		}
		if(axis instanceof EnumerationAxis) {
			EnumerationAxis<Object> ea = (EnumerationAxis<Object>)axis;
			int segments = style.isUseIndividualAxisSegments()
					? style.getAxisStyle(axis).getSegments() : style.getAxisSegments();
			int index = Math.max(0, ea.indexOfEnumerationMember(data.getAxisValue(ea)));
			ratio = (double)index / (double)segments;
		}
		Point outer = axesEnds.get(axis);
		int deltaX = (int)(ratio * (outer.x - center.x));
		int deltaY = (int)(ratio * (outer.y - center.y));
		return new Point(center.x + deltaX, center.y + deltaY);
	}

	/** Draws the axes of the spider chart. */
	private Map<AxisBase, Point> drawAxes(GCStateManager gcState) {
		List<AxisBase> axes = chart.getAxes();
		int size = axes.size();
		double segmentDegree = 360.0 / size;
		double angleDegree = 90.0;

		int halfWidth = getWidth() / 2;
		int halfHeight = getHeight() / 2;
		int centerX = getX() + halfWidth;
		int centerY = getY() + halfHeight;

		for(int i = 0; i < size; i++) {
			AxisBase axis = axes.get(i);
			Point outer = axesEnds.get(axis);
			// draw line
			AxisStyle aStyle = style.getAxisStyle(axis);
			aStyle.getLineStyle().draw(gcState, centerX, centerY, outer.x, outer.y);
			// draw label
			drawAxisLabel(gcState, axis, aStyle, centerX, centerY, outer.x, outer.y);

			if(aStyle.isUseSegmentIndicators()) {
				drawSegmentIndicators(gcState, centerX, centerY, axis, outer, aStyle);
			}
			angleDegree = (angleDegree + segmentDegree) % 360.0;
		}
		return axesEnds;
	}

	/** Draws the indicators of the given axis. */
	private void drawSegmentIndicators(GCStateManager gcState, int centerX, int centerY,
			AxisBase axis, Point outer, AxisStyle aStyle) {
		// draw segment indicators
		int segments = style.isUseIndividualAxisSegments() ? aStyle.getSegments()
				: style.getAxisSegments();
		for(int s = 1; s <= segments; s++) {
			String lbl = "";
			if(axis instanceof DoubleAxis) {
				double ratio = (double)s / (double)segments;
				Double value = ((DoubleAxis)axis).getAxisValue(ratio);
				lbl = aStyle.getDecimalFormat().format(value);
			} else if(axis instanceof EnumerationAxis) {
				Object value = ((EnumerationAxis<?>)axis).getEnumerationMemberAtIndex(s);
				if(value != null) {
					lbl = value.toString();
				}
			}
			Point segmentPoint =
					getSegmentPoint(segments, s, outer.x - center.x, outer.y - center.y);
			drawSegmentIndicator(gcState, lbl, aStyle, centerX, centerY, segmentPoint);
		}
	}

	/** Draws the indicator on the axis. */
	private void drawSegmentIndicator(GCStateManager gcState, String lbl, AxisStyle axisStyle,
			int centerX, int centerY, Point segment) {
		FontStyle segmentStyle = axisStyle.getSegmentStyle();
		gcState.fontStoreAndCreate(segmentStyle.getFontData());
		gcState.foregroundStoreAndCreate(segmentStyle.getRGBColor());

		GC gc = gcState.GC();
		// segment indicator labels are always to the left of the axis
		int posX = centerX + segment.x - gc.textExtent(lbl).x - 4;
		int posY = centerY + segment.y + 4;
		gc.drawText(lbl, posX, posY, true);

		gcState.foregroundRestoreAndDispose().fontRestoreAndDispose();
	}

	/** Draws the axis label at the correct position. */
	private void drawAxisLabel(GCStateManager gcState, AxisBase axis, AxisStyle style, int centerX,
			int centerY, int outerX, int outerY) {
		String axisLabel = axis.getName();
		FontData fontData = style.getLabelStyle().getFontData();

		gcState.fontStoreAndCreate(fontData);
		gcState.foregroundStoreAndCreate(style.getLabelStyle().getRGBColor());

		int textExtent = gcState.GC().textExtent(axisLabel).x;
		if(outerX > centerX) {
			outerX += 5;
		} else if(outerX < centerX) {
			outerX = outerX - 5 - textExtent;
		} else {
			outerX = outerX - textExtent / 2;
		}
		if(outerY > centerY) {
			outerY = outerY + 5;
		} else if(outerY < centerY) {
			outerY = outerY - 5 - fontData.getHeight();
		} else {
			outerY = outerY - fontData.getHeight() / 2;
		}
		gcState.GC().drawString(axisLabel, outerX, outerY, true);

		gcState.foregroundRestoreAndDispose().fontRestoreAndDispose();
	}
}
