/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IMVCBundle;

import javafx.collections.ObservableList;
import javafx.scene.Group;
import javafx.scene.Node;
import javafx.scene.transform.Scale;

/** {@link Group} of layers for the {@link DiagramViewer}. */
public final class DiagramLayers extends Group {
	/** The diagram viewer of this layer group. */
	private final DiagramViewer viewer;
	/** The bottom layer. */
	private final Layer bottomLayer = new Layer();
	/** The content node layer. */
	private final Layer contentLayer = new Layer();
	/** The link node layer. */
	private final Layer linkLayer = new Layer();
	/** The anchorage node layer. */
	private final Layer anchorageLayer = new Layer();
	/** The text node layer. */
	private final Layer textLayer = new Layer();
	/** The feedback layer for visual feedback. */
	private final Layer visualFeedbackLayer = new Layer();
	/** The feedback layer for interaction with content objects. */
	private final Layer contentInteractionLayer = new Layer();
	/** The feedback layer for interaction with anchorage objects. */
	private final Layer anchorageInteractionLayer = new Layer();
	/** The feedback layer for interaction with link objects. */
	private final Layer linkInteractionLayer = new Layer();
	/** The top layer. */
	private final Layer topLayer = new Layer();
	/** The current mouse drag state associated with this bundle's visual. */
	private final MouseState mouseState;
	/** The scale to be applied for zooming the content of the layers. */
	private Scale scale = new Scale();

	/** Constructor. */
	public DiagramLayers(DiagramViewer viewer) {
		this.viewer = viewer;
		ObservableList<Node> c = getChildren();
		c.add(bottomLayer);
		c.add(contentLayer);
		c.add(anchorageLayer);
		c.add(linkLayer);
		c.add(visualFeedbackLayer);
		c.add(textLayer);
		c.add(contentInteractionLayer);
		c.add(anchorageInteractionLayer);
		c.add(linkInteractionLayer);
		c.add(topLayer);
		mouseState = new MouseState(viewer);
	}

	/** Returns mouseState. */
	public MouseState getMouseState() {
		return mouseState;
	}

	/** Returns the bottom layer. */
	public ILayer getBottomLayer() {
		return bottomLayer;
	}

	/** Returns the content layer. */
	public ILayer getContentLayer() {
		return contentLayer;
	}

	/** Returns the anchorage lLayer. */
	public ILayer getAnchorageLayer() {
		return anchorageLayer;
	}

	/** Returns the link layer. */
	public ILayer getLinkLayer() {
		return linkLayer;
	}

	/** Returns the visual feedback layer. */
	public ILayer getVisualFeedbackLayer() {
		return visualFeedbackLayer;
	}

	/** Returns the text layer. */
	public ILayer getTextLayer() {
		return textLayer;
	}

	/**
	 * Returns the layer for (usually invisible) nodes used for user interactions with content
	 * nodes.
	 */
	public ILayer getContentInteractionLayer() {
		return contentInteractionLayer;
	}

	/**
	 * Returns the layer for (usually invisible) nodes used for user interactions with anchorage
	 * nodes.
	 */
	public ILayer getAnchorageInteractionLayer() {
		return anchorageInteractionLayer;
	}

	/**
	 * Returns the layer for (usually invisible) nodes used for user interactions with link nodes.
	 */
	public ILayer getLinkInteractionLayer() {
		return linkInteractionLayer;
	}

	/** Returns the top layer. */
	public ILayer getTopLayer() {
		return topLayer;
	}

	/** Returns whether the given node is an interactive feedback node. */
	public boolean isInteractionFeedback(Node node) {
		return node.getParent() == contentInteractionLayer;
	}

	/** Clears all nodes from the layers. */
	public void clearLayers() {
		bottomLayer.clear();
		contentLayer.clear();
		anchorageLayer.clear();
		linkLayer.clear();
		visualFeedbackLayer.clear();
		textLayer.clear();
		contentInteractionLayer.clear();
		anchorageInteractionLayer.clear();
		linkInteractionLayer.clear();
		topLayer.clear();
	}

	/** Registers the scroll listener and the zoom scale transformation. */
	private void registerScrollListenerAndZoomScale(Node node) {
		node.setOnScroll(viewer.getScrollingHandler());
		node.getTransforms().add(scale);
	}

	/** Unregisters the scroll listener and the zoom scale transformation. */
	private void unregisterScrollListenerAndZoomScale(Node node) {
		node.setOnScroll(null);
		node.getTransforms().remove(scale);
	}

	/** Scales all nodes in each layer. */
	/* package */ void setScale(double factor) {
		scale.setX(factor);
		scale.setY(factor);
		scale.setZ(factor);
	}

	/** Interface for layers. */
	public static interface ILayer {
		/**
		 * Adds the given bundle's node to this layer. Note that the given node's parent must be
		 * null. In other words, a node must first be explicitly removed from its former layer and
		 * then added to the new one.
		 */
		void add(Node node, IMVCBundle bundle);

		/** Removes the given node from this layer. */
		void remove(Node node);

		/** Clears this layer removing all nodes. */
		void clear();
	}

	/** Implementation of layers. */
	private class Layer extends Group implements ILayer {
		/** {@inheritDoc} */
		@Override
		public void add(Node node, IMVCBundle bundle) {
			if(node == null || bundle == null || node.getParent() != null) {
				return;
			}
			getChildren().add(node);
			registerScrollListenerAndZoomScale(node);
			mouseState.registerMouseListeners(node, bundle);
		}

		/** {@inheritDoc} */
		@Override
		public void remove(Node node) {
			if(node == null || node.getParent() != this) {
				return;
			}
			mouseState.unregisterMouseListeners(node);
			unregisterScrollListenerAndZoomScale(node);
			getChildren().remove(node);
		}

		/** {@inheritDoc} */
		@Override
		public void clear() {
			for(Node node : getChildren()) {
				mouseState.unregisterMouseListeners(node);
				unregisterScrollListenerAndZoomScale(node);
			}
			getChildren().clear();
		}
	}
}
