/*******************************************************************************
 * Copyright (c) 2019 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.base;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewer;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.FeedbackChange;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IDragController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IContentAnchorageMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IMVCBundle;

/**
 * Base class providing dragging implementation for visuals which move and resize in the
 * {@link DiagramViewer} and which have movable anchorages attached. Sub-classes must implement
 * {@link #resize(FeedbackChange)} and {@link #move(FeedbackChange)} in order to change the
 * underlying model with the respective coordinate changes. Furthermore, sub-classes must implement
 * {@link #createAnchorageMoveController(IContentAnchorageMVCBundle)} and create
 * {@link IDragController}s for the given anchorages.
 */
public abstract class AnchorageContentControllerBase extends ResizableContentControllerBase {
	/** Constructor. */
	public AnchorageContentControllerBase(IContentMVCBundle mvcb) {
		super(mvcb);
	}

	/** Stores the anchorage move controllers. */
	private final Map<IContentAnchorageMVCBundle, IDragController> moveAnchorageControllerMap =
			new HashMap<IContentAnchorageMVCBundle, IDragController>();

	/**
	 * Returns an {@link IDragController} responsible for positioning the anchorage visual during
	 * the move feedback and for committing the new position to the model.
	 */
	protected final IDragController
			getAnchorageMoveController(IContentAnchorageMVCBundle anchorage) {
		if(!canMoveAnchorage(anchorage)) {
			return null;
		}
		IDragController ctrl = moveAnchorageControllerMap.get(anchorage);
		if(ctrl == null) {
			ctrl = createAnchorageMoveController(anchorage);
			moveAnchorageControllerMap.put(anchorage, ctrl);
		}
		return ctrl;
	}

	/**
	 * Returns whether the given content anchorage should be movable.
	 * 
	 * @param anchorage
	 *            the {@link IMVCBundle} of the anchorage to be moved
	 * @return whether the anchorage can be moved
	 */
	protected boolean canMoveAnchorage(IContentAnchorageMVCBundle anchorage) {
		return true;
	}

	/**
	 * Creates the move controller for the given anchorage.
	 * 
	 * @param anchorage
	 *            the {@link IMVCBundle} of the anchorage to be moved
	 * @return a drag controller for moving the anchorage
	 */
	protected abstract IDragController
			createAnchorageMoveController(IContentAnchorageMVCBundle anchorage);
}
