/*******************************************************************************
 * Copyright (c) 2017, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.base;

import java.util.LinkedList;
import java.util.List;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewerFeatures;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.change.Change;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IClickController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IDragController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IMVCBundle;

import javafx.scene.Node;
import javafx.scene.control.Menu;
import javafx.scene.control.MenuItem;

/**
 * Default implementation for diagram controller, which provides a {@code Display...} context menu
 * with the options to enable and disable link highlighting and interactive area shading.
 */
public class DefaultDiagramController extends ControllerBase {
	/** Flag for ignoring mouse click event after dragging gesture. */
	private boolean ignoreNextMouseClick = false;

	/** Constructor. */
	public DefaultDiagramController(IMVCBundle mvcb) {
		super(mvcb);
	}

	/** {@inheritDoc} */
	@Override
	public List<MenuItem> contextMenuContributions(Node node, DiagramCoordinate diagramLocation) {
		Menu menu = new Menu("Display ...");
		DiagramViewerFeatures features = getViewer().getFeatures();
		// link highlighting
		if(features.isLinkHighlightingEnabled()) {
			MenuItem disableLinkHighlighting = new MenuItem("Disable Link Highlighting");
			disableLinkHighlighting.setOnAction(evt -> {
				features.setLinkHighlightingEnabled(false);
			});
			menu.getItems().add(disableLinkHighlighting);
		} else {
			MenuItem enableLinkHighlighting = new MenuItem("Enable Link Highlighting");
			enableLinkHighlighting.setOnAction(evt -> {
				features.setLinkHighlightingEnabled(true);
			});
			menu.getItems().add(enableLinkHighlighting);
		}
		// interaction area shading
		if(features.isInteractionAreaShadingEnabled()) {
			MenuItem disableInteractionShading = new MenuItem("Disable Interactive Area Shading");
			disableInteractionShading.setOnAction(evt -> {
				features.setInteractionAreaShadingEnabled(false);
			});
			menu.getItems().add(disableInteractionShading);
		} else {
			MenuItem enableInteractionShading = new MenuItem("Enable Interactive Area Shading");
			enableInteractionShading.setOnAction(evt -> {
				features.setInteractionAreaShadingEnabled(true);
			});
			menu.getItems().add(enableInteractionShading);
		}

		List<MenuItem> menuList = new LinkedList<>();
		menuList.add(menu);
		return menuList;
	}

	/** {@inheritDoc} */
	@Override
	public IDragController getDragController(EDragGesture gesture, Node node,
			DiagramCoordinate diagramLocation) {
		if(gesture == EDragGesture.SELECTION) {
			return dragSelectionRectangleController;
		}
		return super.getDragController(gesture, node, diagramLocation);
	}

	/** {@link IDragController} for the selection rectangle. */
	private final IDragController dragSelectionRectangleController = new DragControllerBase() {
		@Override
		protected void safeDragStarted(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			getViewer().startSelectionFeedback(node, diagramLocation);
			ignoreNextMouseClick = true;
		}

		@Override
		protected void safeDragInProgress(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			getViewer().updateSelectionFeedback(node, diagramLocation);
		}

		@Override
		protected Change safeDragCompleted(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			getViewer().terminateSelectionFeedback(node, diagramLocation);
			return null;
		}
	};

	/** {@inheritDoc} */
	@Override
	public IClickController getClickController(Node node, DiagramCoordinate diagramLocation) {
		if(ignoreNextMouseClick) {
			ignoreNextMouseClick = false;
			return null;
		}
		return super.getClickController(node, diagramLocation);
	}
}
