/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.elliptic;

import static java.lang.Math.atan2;
import static java.lang.Math.cos;
import static java.lang.Math.sin;
import static java.util.Objects.requireNonNull;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;

import javafx.geometry.Dimension2D;
import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;

/** This class encapsulates locations on the border of an ellipse defined by the angular value. */
public final class EllipticBorderLocation {
	/** The angle in degrees. */
	private final double angleInDegree;
	/** The radiuses of the ellipse. */
	private final Dimension2D radiuses;
	/**
	 * The correction values used to compute the final location (usually half the size of the object
	 * located on the border so that it looks centered.
	 */
	private final Dimension2D correction;

	/** Constructor. */
	public EllipticBorderLocation(double angleInDegree, Dimension2D radiuses,
			Dimension2D correction) {
		this.angleInDegree = angleInDegree;
		this.radiuses = requireNonNull(radiuses);
		this.correction = requireNonNull(correction);
	}

	/** Returns the angle in degree. */
	public double getAngleInDegree() {
		return angleInDegree;
	}

	/** Returns the radiuses. */
	public Dimension2D getRadiuses() {
		return radiuses;
	}

	/**
	 * Returns the location computed from this {@link EllipticBorderLocation} relative to the
	 * ellipse center.
	 */
	public DiagramCoordinate getLocation() {
		double rw = radiuses.getWidth();
		double rh = radiuses.getHeight();
		double angle = Math.PI * angleInDegree / 180;
		double x = rw - correction.getWidth() / 2 + rw * cos(angle);
		double y = rh - correction.getHeight() / 2 - rh * sin(angle);
		return new DiagramCoordinate(x, y);
	}

	/**
	 * Returns the closest point on the ellipse bounds inside the given rectangle when pointing at
	 * the given location.
	 */
	public static final EllipticBorderLocation getClosestLocationOnBounds(Point2D point,
			Rectangle2D ellipse, Dimension2D size) {
		double w2 = ellipse.getWidth() / 2;
		double h2 = ellipse.getHeight() / 2;
		double dx = point.getX() - w2;
		double dy = h2 - point.getY();

		double angle = atan2(dy, dx) * 180.0 / Math.PI;
		Dimension2D radiuses = new Dimension2D(w2, h2);
		return new EllipticBorderLocation(angle, radiuses, size);
	}
}
