/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.elliptic;

import static java.lang.Math.atan2;
import static java.lang.Math.toDegrees;
import static javafx.scene.paint.Color.DARKORANGE;
import static javafx.scene.paint.Color.ORANGE;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.elliptic.EllipticBorderLocation.getClosestLocationOnBounds;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramLayers;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.model.layout.IAngleLayout;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.IContentAnchorageVisual;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.base.ContentVisualBase;

import javafx.geometry.Bounds;
import javafx.geometry.Dimension2D;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.paint.Paint;
import javafx.scene.shape.Ellipse;

/** Base class for {@link ContentVisualBase content visuals} depicted by ellipses. */
public abstract class EllipticContentVisualBase extends ContentVisualBase {
	/** Constructor. */
	public EllipticContentVisualBase(IContentMVCBundle mvcb) {
		super(mvcb, new Ellipse(), new Ellipse());
	}

	/** {@inheritDoc} */
	@Override
	public void updateNodes(DiagramLayers layers) {
		super.updateNodes(layers);
		updateNodeProperties();
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getAnchorageLocation(IContentAnchorageVisual visual) {
		Rectangle2D pb = getCurrentBounds();
		Dimension2D ell = new Dimension2D(pb.getWidth() / 2, pb.getHeight() / 2);
		Dimension2D dim = visual.getDimensions();
		double angleDeg = visual.getLayout(IAngleLayout.class).getAngleInDegree();
		EllipticBorderLocation ebl = new EllipticBorderLocation(angleDeg, ell, dim);
		return ebl.getLocation();
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getLinkAnchorage(DiagramCoordinate indication) {
		Rectangle2D pb = getCurrentBounds();
		EllipticBorderLocation ebl = getClosestLocationOnBounds(
				indication.add(-pb.getMinX(), -pb.getMinY()), pb, new Dimension2D(1, 1));
		return ebl.getLocation().add(pb.getMinX(), pb.getMinY());
	}

	/** {@inheritDoc} */
	@Override
	protected Paint getBorderColor() {
		return DARKORANGE.darker();
	}

	/** {@inheritDoc} */
	@Override
	protected Paint getFillColor() {
		return ORANGE.brighter();
	}

	/** Updates the properties of the visual nodes. */
	private void updateNodeProperties() {
		Rectangle2D bounds = getCurrentBounds();
		double w2 = bounds.getWidth() / 2;
		double h2 = bounds.getHeight() / 2;
		Ellipse ellipse = getVisualEllipse();
		ellipse.setCenterX(bounds.getMinX() + w2);
		ellipse.setCenterY(bounds.getMinY() + h2);
		ellipse.setRadiusX(w2);
		ellipse.setRadiusY(h2);
		ellipse.setOpacity(getOpacity());
		ellipse.setStroke(getBorderColor());
		ellipse.setStrokeWidth(getBorderWidth());
		ellipse.setStrokeType(getBorderType());
		ellipse.setFill(getFillColor());

		double lnk = getHitAreaStartLinkSize();
		double res = getHitAreaResizeSize();
		// wild cast works: see constructor
		Ellipse hitShape = getHitAreaEllipse();
		hitShape.setCenterX(bounds.getMinX() + w2);
		hitShape.setCenterY(bounds.getMinY() + h2);
		hitShape.setRadiusX(w2 + lnk + res / 2);
		hitShape.setRadiusY(h2 + lnk + res / 2);
		Paint interactionColor = getInteractionColor();
		hitShape.setStroke(interactionColor);
		hitShape.setFill(interactionColor);
	}

	/** {@inheritDoc} */
	@Override
	protected EDragGesture dragGestureHitTest(Node node, DiagramCoordinate diagramLocation) {
		if(node == getHitAreaShape()) {
			Bounds bounds = node.getBoundsInLocal();
			double w = bounds.getWidth();
			double h = bounds.getHeight();

			double x = diagramLocation.getLocalX(node) - w / 2;
			double y = diagramLocation.getLocalY(node) - h / 2;
			double res = getHitAreaResizeSize();
			double lnk = getHitAreaStartLinkSize();
			// check for move area
			double a = w / 2 - res - lnk;
			double b = h / 2 - res - lnk;
			double hitMoveArea = (x * x / (a * a) + y * y / (b * b));
			if(hitMoveArea <= 1.0) {
				return EDragGesture.MOVE;
			}
			// check for resize area
			a = w / 2 - lnk;
			b = h / 2 - lnk;
			double hitResizeArea = (x * x / (a * a) + y * y / (b * b));
			if(hitResizeArea <= 1.0) {
				// check angle
				double angle = (360.0 + toDegrees(atan2(-y, x))) % 360.0;
				if(angle <= 45.0 || angle >= 345.0) {
					return EDragGesture.RESIZE_H;
				}
				if(angle >= 300.0) {
					return EDragGesture.RESIZE_VH;
				}
				if(angle >= 225.0) {
					return EDragGesture.RESIZE_V;
				}
				return EDragGesture.NONE;
			}
			// else startlink
			return EDragGesture.NEW_LINK;
		}
		if(node == getVisualEllipse() || node == text) {
			return EDragGesture.MOVE;
		}
		return EDragGesture.NONE;
	}

	/**
	 * Returns the extension of the hit area in pixel used to trigger link creation. This number
	 * specifies how many pixels beyond the visible border are used to trigger link creation by a
	 * mouse drag gesture.
	 */
	protected double getHitAreaStartLinkSize() {
		return getViewer().getFeatures().getMaximumSpacing();
	}

	/** Returns the size of the hit area in pixel used to resize the content visual. */
	protected double getHitAreaResizeSize() {
		return getViewer().getFeatures().getMaximumSpacing();
	}

	/** Returns the ellipse visual. */
	protected final Ellipse getVisualEllipse() {
		// wild cast works: see constructor
		return (Ellipse)getVisualShape();
	}

	/** Returns the ellipse of the hit area. */
	protected final Ellipse getHitAreaEllipse() {
		return (Ellipse)getHitAreaShape();
	}
}
