/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.rectangular;

import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.*;

import java.util.LinkedList;
import java.util.List;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramLayers;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramLayers.ILayer;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.FeedbackChange;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IAnchorageMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.impl.LinkMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.base.LinkVisualBase;

import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.paint.Color;
import javafx.scene.paint.Paint;

/** Base class for {@link LinkVisualBase link visuals} with straight lines. */
public abstract class LineLinkVisualBase extends LinkVisualBase {
	/** Stores the visual segments. */
	private final List<LineSegment> segments = new LinkedList<>();
	/** Flag storing if the lines have been added. */
	private boolean linesAddedToSceneGraph = false;

	/** Constructor. */
	public LineLinkVisualBase(ILinkMVCBundle mvcb) {
		super(mvcb);
	}

	/** Creates the line segments and adds them to the scene graph. */
	private void addLineSegments(DiagramLayers layers) {
		Object bpModel = null;
		double feedbackSize = getFeedbackMarkerSize();
		double fs2 = feedbackSize / 2;
		// start point
		Point2D startAnchorPoint = getFeedbackChangeForBendPoint(START_OF_LINK_BEND_POINT_INDEX)
				.applyToPoint(getStartAnchorPoint());
		double sx = startAnchorPoint.getX();
		double sy = startAnchorPoint.getY();
		startFeedbackHandle.setX(sx - fs2);
		startFeedbackHandle.setY(sy - fs2);
		startFeedbackHandle.setWidth(feedbackSize);
		startFeedbackHandle.setHeight(feedbackSize);
		// handle bend points
		int pts = getNumberOfBendPoints();
		for(int i = 0; i < pts; i++) {
			Point2D bp = getBendPointLocation(i);
			FeedbackChange chg = getFeedbackChangeForBendPoint(i);
			double nx = bp.getX() + chg.getDeltaX();
			double ny = bp.getY() + chg.getDeltaY();
			segments.add(new LineSegment(bpModel, sx, sy, nx, ny, getInvisibleSelectionLineWidth(),
					feedbackSize, false, false, 0, getLabelText(i, pts + 1)));
			sx = nx;
			sy = ny;
			bpModel = getBendPointModel(i);
		}
		// end point
		Point2D endPoint = getFeedbackChangeForBendPoint(END_OF_LINK_BEND_POINT_INDEX)
				.applyToPoint(getEndAnchorPoint());
		segments.add(new LineSegment(bpModel, sx, sy, endPoint.getX(), endPoint.getY(),
				getInvisibleSelectionLineWidth(), 0, showArrowOnLastSegment(), useLineArrow(),
				getArrowLength(), getLabelText(pts, pts + 1)));
		endFeedbackHandle.setX(endPoint.getX() - fs2);
		endFeedbackHandle.setY(endPoint.getY() - fs2);
		endFeedbackHandle.setWidth(feedbackSize);
		endFeedbackHandle.setHeight(feedbackSize);

		ILinkMVCBundle bundle = getLinkBundle();
		int numberOfSegs = segments.size();
		for(int i = 0; i < numberOfSegs; i++) {
			LineSegment ls = segments.get(i);
			ls.addLinkNodes(layers, bundle, getLabelText(i, numberOfSegs));
		}
		linesAddedToSceneGraph = true;
	}

	/** {@inheritDoc} */
	@Override
	public void removeAllVisuals(DiagramLayers layers) {
		super.removeAllVisuals(layers);
		removeLineSegments(layers);
	}

	/** Removes the line segments from the scene graph. */
	private void removeLineSegments(DiagramLayers layers) {
		for(LineSegment ls : segments) {
			ls.removeLinkNodes(layers);
			ls.removeFeedbackNodes(layers);
		}
		ILayer linkInteractionLayer = layers.getLinkInteractionLayer();
		linkInteractionLayer.remove(startFeedbackHandle);
		linkInteractionLayer.remove(endFeedbackHandle);
		segments.clear();
		linesAddedToSceneGraph = false;
	}

	/** {@inheritDoc} */
	@Override
	public void updateNodes(DiagramLayers layers) {
		super.updateNodes(layers);
		if(!enableVisual()) {
			if(linesAddedToSceneGraph) {
				removeLineSegments(layers);
			}
			return;
		}

		int pts = getNumberOfBendPoints();
		int expectedSegments = 1 + pts;
		if(!linesAddedToSceneGraph) {
			addLineSegments(layers);
		}
		if(segments.size() != expectedSegments) {
			removeLineSegments(layers);
			addLineSegments(layers);
			if(getViewer().getSelection().getPrimarySelection() == getLinkBundle()) {
				if(getViewer().hasFocus()) {
					applySelectedFocusedEffect(layers);
				} else {
					applySelectedNotFocusedEffect(layers);
				}
			}
			return;
		}
		int segIndex = 0;
		double feedbackSize = getFeedbackMarkerSize();
		double fs2 = feedbackSize / 2;
		// start point
		Point2D sp = getFeedbackChangeForBendPoint(START_OF_LINK_BEND_POINT_INDEX)
				.applyToPoint(getStartAnchorPoint());
		startFeedbackHandle.setX(sp.getX() - fs2);
		startFeedbackHandle.setY(sp.getY() - fs2);
		startFeedbackHandle.setWidth(feedbackSize);
		startFeedbackHandle.setHeight(feedbackSize);
		// handle bend points
		for(int i = 0; i < pts; i++) {
			Point2D ep = getFeedbackChangeForBendPoint(i).applyToPoint(getBendPointLocation(i));
			segments.get(segIndex).update(sp, ep, feedbackSize, 0, getLabelText(i, pts));
			sp = ep;
			segIndex++;
		}
		// end point
		Point2D ep = getFeedbackChangeForBendPoint(END_OF_LINK_BEND_POINT_INDEX)
				.applyToPoint(getEndAnchorPoint());
		segments.get(segIndex).update(sp, ep, 0, getArrowLength(), getLabelText(pts, pts));
		endFeedbackHandle.setX(ep.getX() - fs2);
		endFeedbackHandle.setY(ep.getY() - fs2);
		endFeedbackHandle.setWidth(feedbackSize);
		endFeedbackHandle.setHeight(feedbackSize);

		Paint color = getLineColor();
		IMVCBundle mvcBundle = getMVCBundle();
		if(mvcBundle.hasTag(HIGHLIGHT_INCOMING_LINK_TAG)) {
			color = getHighlightIncomingLinkColor();
		} else if(mvcBundle.hasTag(HIGHLIGHT_OUTGOING_LINK_TAG)) {
			color = getHighlightOutgoingLinkColor();
		} else if(mvcBundle.hasTag(PRIMARY_SELECTION_TAG)) {
			if(mvcBundle.hasTag(FOCUS_TAG)) {
				color = getSelectionFocusedColor();
			} else {
				color = getSelectionNotFocusedColor();
			}
		}
		for(LineSegment seg : segments) {
			seg.setStrokeColor(color);
		}
	}

	/** {@inheritDoc} */
	@Override
	public boolean enableVisual() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean enableInteraction() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getStartAnchorPoint() {
		IAnchorageMVCBundle start = getLinkBundle().getStartAnchorage();
		Rectangle2D sb = start.getVisual().getCurrentBounds();

		DiagramCoordinate target;
		int pts = getNumberOfBendPoints();
		if(pts == 0) {
			IAnchorageMVCBundle end = getLinkBundle().getEndAnchorage();
			Rectangle2D eb = end.getVisual().getCurrentBounds();
			target = new DiagramCoordinate(eb.getMinX() + eb.getWidth() / 2,
					eb.getMinY() + eb.getHeight() / 2);
		} else {
			target = getCurrentBendPointLocation(0);
		}
		return getStartAnchorLocation(sb, target);
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getEndAnchorPoint() {
		IAnchorageMVCBundle end = getLinkBundle().getEndAnchorage();
		Rectangle2D eb = end.getVisual().getCurrentBounds();

		DiagramCoordinate target;
		int pts = getNumberOfBendPoints();
		if(pts == 0) {
			IAnchorageMVCBundle start = getLinkBundle().getStartAnchorage();
			Rectangle2D sb = start.getVisual().getCurrentBounds();
			target = new DiagramCoordinate(sb.getMinX() + sb.getWidth() / 2,
					sb.getMinY() + sb.getHeight() / 2);
		} else {
			target = getCurrentBendPointLocation(pts - 1);
		}
		return getEndAnchorLocation(eb, target);
	}

	/** {@inheritDoc} */
	@Override
	public void applySelectedFocusedEffect(DiagramLayers layers) {
		ILinkMVCBundle bundle = getLinkBundle();
		Paint interactionColor = getInteractionColor();
		Paint selectionFocusedColor = getSelectionFocusedColor();
		for(LineSegment seg : segments) {
			seg.setStrokeColor(selectionFocusedColor);
			seg.setClickableLinesColor(interactionColor);
			seg.addFeedbackNodes(layers, bundle);
		}
		ILayer linkInteractionLayer = layers.getLinkInteractionLayer();
		linkInteractionLayer.add(startFeedbackHandle, bundle);
		linkInteractionLayer.add(endFeedbackHandle, bundle);
	}

	/** {@inheritDoc} */
	@Override
	public void applySelectedNotFocusedEffect(DiagramLayers layers) {
		removeFeedback(layers, getSelectionNotFocusedColor());
	}

	/** {@inheritDoc} */
	@Override
	protected void applyNotSelectedEffect(DiagramLayers layers) {
		removeFeedback(layers, getLineColor());
	}

	/** Removes the feedback handles and sets the link to the given color. */
	private void removeFeedback(DiagramLayers layers, Paint lineColor) {
		Paint interactionColor = getInteractionColor();
		for(LineSegment seg : segments) {
			seg.setStrokeColor(lineColor);
			seg.setClickableLinesColor(interactionColor);
			seg.removeFeedbackNodes(layers);
		}
		ILayer linkInteractionLayer = layers.getLinkInteractionLayer();
		linkInteractionLayer.remove(startFeedbackHandle);
		linkInteractionLayer.remove(endFeedbackHandle);
	}

	/**
	 * Returns the anchor location w.r.t. the interface node bonds and the given target location
	 * (e.g. the first bend-point or the end interface node center point).
	 */
	protected abstract DiagramCoordinate getStartAnchorLocation(Rectangle2D ifcBounds,
			DiagramCoordinate target);

	/**
	 * Returns the anchor location w.r.t. the interface node bonds and the given target location
	 * (e.g. the last bend-point or the start interface node center point).
	 */
	protected abstract DiagramCoordinate getEndAnchorLocation(Rectangle2D ifcBounds,
			DiagramCoordinate target);

	/** Returns the color to be used when the connection is selected and focused. */
	protected Paint getSelectionFocusedColor() {
		return Color.RED;
	}

	/** Returns the color to be used when the connection is selected, but not focused. */
	protected Paint getSelectionNotFocusedColor() {
		return Color.ORANGE;
	}

	/** Returns the color to be used when the connection is not selected. */
	protected Paint getLineColor() {
		return Color.BLACK;
	}

	/**
	 * Returns the point of the bend-point at the given index altered by its current feedback
	 * change.
	 */
	private DiagramCoordinate getCurrentBendPointLocation(int bendPointIndex) {
		if(bendPointIndex >= 0 && bendPointIndex < getNumberOfBendPoints()) {
			Point2D bendPoint = getBendPointLocation(bendPointIndex);
			FeedbackChange chg = getFeedbackChangeForBendPoint(bendPointIndex);
			double x = bendPoint.getX() + chg.getDeltaX();
			double y = bendPoint.getY() + chg.getDeltaY();
			return new DiagramCoordinate(x, y);
		}
		throw new IndexOutOfBoundsException("No bend-point exists at index " + bendPointIndex);
	}

	/**
	 * Returns the model of the bend-point at the given index. Each bend-point must be identified by
	 * a unique object, which must not change as long as the bend-point exists.
	 */
	protected abstract Object getBendPointModel(int i);

	/** Returns the number of bend-points. */
	protected abstract int getNumberOfBendPoints();

	/** Returns whether the last line segment should show an arrow head marker. */
	protected abstract boolean showArrowOnLastSegment();

	/** Returns the type of the arrow to be used (line vs. filled). The default is filled. */
	protected boolean useLineArrow() {
		return false;
	}

	/** Returns the length of the arrow. */
	protected double getArrowLength() {
		return 10;
	}

	/** Returns the width of the invisible selection line. */
	protected abstract double getInvisibleSelectionLineWidth();

	/**
	 * Returns the label text.
	 * 
	 * @param currentSegment
	 *            the current segment
	 * @param lastSegment
	 *            the last index of a segment
	 * @return the label text for the current segment or null for no label
	 */
	protected String getLabelText(int currentSegment, int lastSegment) {
		return null;
	}

	/** Returns the {@link LinkMVCBundle}. */
	private LinkMVCBundle getLinkBundle() {
		return (LinkMVCBundle)getMVCBundle();
	}

	/** {@inheritDoc} */
	@Override
	public Rectangle2D getCurrentBounds() {
		// connections do not need bounds
		return Rectangle2D.EMPTY;
	}

	/** {@inheritDoc} */
	@Override
	public int getBendPointIndex(Node node) {
		int superResult = super.getBendPointIndex(node);
		if(superResult != NO_BEND_POINT_INDEX) {
			return superResult;
		}
		for(int i = 0; i < segments.size(); i++) {
			LineSegment ls = segments.get(i);
			if(node == ls.getFeedbackHandle() || node == ls.getClickableLine() ||
					node == ls.getVisibleLine()) {
				return i;
			}
		}
		return NO_BEND_POINT_INDEX;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isBendPointHandle(Node node) {
		for(LineSegment ls : segments) {
			if(node == ls.getFeedbackHandle()) {
				return true;
			}
			if(node == ls.getClickableLine() || node == ls.getVisibleLine()) {
				return false;
			}
		}
		return super.isBendPointHandle(node);
	}

	/** {@inheritDoc} */
	@Override
	public void requestFocus() {
		if(!segments.isEmpty()) {
			// segments list tested for emptiness => get(0) works
			segments.get(0).getVisibleLine().requestFocus();
		}
	}

	/** {@inheritDoc} */
	@Override
	public EDragGesture getDragGesture(Node node, DiagramCoordinate diagramLocation) {
		if(node == startFeedbackHandle || node == endFeedbackHandle) {
			return EDragGesture.RECONNECT_LINK;
		}
		for(LineSegment ls : segments) {
			if(node == ls.getFeedbackHandle()) {
				return EDragGesture.MOVE_BENDPOINT;
			}
			if(node == ls.getClickableLine()) {
				return EDragGesture.NEW_BENDPOINT;
			}
		}
		return EDragGesture.NONE;
	}

	/** {@inheritDoc} */
	@Override
	public void enableFeedback() {
		super.enableFeedback();
		for(LineSegment ls : segments) {
			ls.setMouseTransparent(false);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void disableFeedback() {
		super.disableFeedback();
		for(LineSegment ls : segments) {
			ls.setMouseTransparent(true);
		}
	}
}
